<?php

namespace Drupal\config_batch_export\Controller;

use Drupal\config\Controller\ConfigController;
use Drupal\Core\Archiver\ArchiveTar;
use Drupal\Core\Serialization\Yaml;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;
use Drupal\file\Entity\File;
use Drupal\file\FileInterface;
use Drupal\system\FileDownloadController;
use Symfony\Component\HttpFoundation\Request;

/**
 * Returns responses for config_batch_export module routes.
 */
class ConfigBatchExportController extends ConfigController {

  use StringTranslationTrait;

  /**
   * How many elements of the array to process per one operation.
   */
  const BATCH_SIZE = 10;

  /**
   * Batch callback.
   */
  public static function callbackBatchJob($batch_data, &$context) {
    $sandbox = &$context['sandbox'];
    if (!isset($sandbox['started'])) {
      /** @var \Drupal\file\FileInterface $file */
      $file = \Drupal::entityTypeManager()->getStorage('file')->load($batch_data['file_id']);

      $sandbox['filename'] = $file->getFileUri();
      $sandbox['started'] = TRUE;
      $sandbox['offset'] = 0;
      $sandbox['total'] = count($batch_data['configs']);
      $sandbox['batch_size'] = static::BATCH_SIZE;
      $context['results']['file_id'] = $batch_data['file_id'];
    }

    $archiver = new ArchiveTar($sandbox['filename'], 'gz');
    $configs = array_slice($batch_data['configs'], $sandbox['offset'], $sandbox['batch_size']);
    $collection_storage_cache = [];
    if (!empty($configs)) {
      $config_factory = \Drupal::configFactory();
      /** @var \Drupal\Core\Config\StorageInterface $target_storage */
      $target_storage = \Drupal::service('config.storage');
      foreach ($configs as $config_definition) {
        if (is_array($config_definition)) {
          // Process collection config.
          $name = $config_definition['name'];
          $collection = $config_definition['collection'];
          $collection_storage = $collection_storage_cache[$collection] ?? ($collection_storage_cache[$collection] = $target_storage->createCollection($collection));
          $archiver->addString(str_replace('.', '/', $collection) . "/$name.yml", Yaml::encode($collection_storage->read($name)));
        }
        else {
          $name = $config_definition;
          $archiver->addString("$name.yml", Yaml::encode($config_factory->get($name)->getRawData()));
        }
      }

      $context['finished'] = $sandbox['offset'] / $sandbox['total'];
    }
    else {
      $context['finished'] = 1;
    }

    // Call destructor.
    unset($archiver);

    $sandbox['offset'] += $sandbox['batch_size'];

    $context['message'] = t('Processed @current out of @total', [
      '@current' => $context['sandbox']['offset'],
      '@total' => $context['sandbox']['total'],
    ]);
  }

  /**
   * Batch finished callback.
   */
  public static function callbackBatchFinished($status, $results) {
    /** @var \Drupal\file\FileInterface $file */
    $file = \Drupal::entityTypeManager()->getStorage('file')->load($results['file_id']);

    $url = Url::fromRoute('config_batch_export.export_download_file', [
      'file' => $file->id(),
    ]);

    \Drupal::messenger()->addMessage(t('Config file created at @created can be downloaded here <a href="@url">here</a>', [
      '@url' => $url->toString(),
      '@created' => date(DateTimeItemInterface::DATE_STORAGE_FORMAT),
    ]));
  }

  /**
   * Downloads a tarball of the site configuration.
   */
  public function downloadExport() {
    /** @var \Drupal\Core\File\FileSystem $file_system */
    $file_system = \Drupal::service('file_system');
    $filepath = 'private://configs.tar.gz';

    // @todo: Use dependency injection for all \Drupal calls.
    // @todo: Use different temporary files with random names.
    // @todo: Control gurbage collection.
    $existing_files = \Drupal::entityTypeManager()->getStorage('file')->loadByProperties(['uri' => $filepath]);
    if (!empty($existing_files)) {
      /** @var FileInterface $file */
      foreach ($existing_files as $file) {
        $file->delete();

        if (file_exists($file->getFileUri())) {
          file_unmanaged_delete($file->getFileUri());
        }
      }
    }

    // Ensure the file is deleted.
    if (file_exists($filepath)) {
      file_unmanaged_delete($filepath);
    }

    $file = File::create([
      'uri' => $filepath,
      'uid' => \Drupal::currentUser()->id(),
    ]);

    $file->setTemporary();
    $file->save();

    $batch_data = [
      'file_id' => $file->id(),
      'configs' => $this->configManager->getConfigFactory()->listAll(),
    ];

    // Get all override data from the remaining collections.
    foreach ($this->targetStorage->getAllCollectionNames() as $collection) {
      $collection_storage = $this->targetStorage->createCollection($collection);
      foreach ($collection_storage->listAll() as $name) {
        $batch_data['configs'][] = [
          'name' => $name,
          'collection' => $collection,
        ];
      }
    }

    $batch = [
      'title' => $this->t('Generation of a configs archive'),
      'operations' => [
        [
          [static::class, 'callbackBatchJob'],
          [$batch_data],
        ],
      ],
      'finished' => [static::class, 'callbackBatchFinished'],
    ];

    batch_set($batch);

    $request =  batch_process(Url::fromRoute('config.export_full')->toString());
    $request->send();
  }

  /**
   * Downloads a file.
   *
   * @param \Drupal\file\FileInterface $file
   *   File object.
   *
   * @return \Symfony\Component\HttpFoundation\BinaryFileResponse
   *   Response.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function downloadExportFile(FileInterface $file) {
    $request = new Request([
      'file' => basename($file->getFileUri()),
    ]);

    // Make the file old to delete it in file_cron.
    $file
      ->setChangedTime(1)
      ->save();

    $scheme = file_uri_scheme($file->getFileUri());
    $download_controller = new FileDownloadController();
    $response = $download_controller->download($request, $scheme);
    return $response;
  }

}
